class PrecisionAnalysis:
    def __init__(self, remote_hostname=None, remote_user=None, remote_scratch=None, job_template=None, vault=None, error_thresholds=None, reduced_precision_level=10, max_running_jobs=10, experiment_name="PrecisionAnalysis", dummy=False, restarting_from_checkpoint=False, assertion_template=None,
                 limit_file=None, analysis_variables=None, forced_variables=None, local_folder=None):
        if dummy is False:
            from AutoRPE.UtilsWorkflow.Communicator import SSH

            self.experiment_name = experiment_name
            self.remote_hostname = remote_hostname
            self.remote_user = remote_user
            self.remote_scratch = remote_scratch
            self.communicator = SSH(user=self.remote_user, host=self.remote_hostname, remote_scratch=self.remote_scratch)
            self.job_template = job_template
            self.vault = vault
            self.error_thresholds = error_thresholds
            self.reduced_precision_level = reduced_precision_level
            self.max_running_jobs = max_running_jobs
            self.restarting_from_checkpoint = restarting_from_checkpoint
            self.assertion_template = assertion_template
            self.analysis_variables = analysis_variables
            self.local_folder = local_folder
            if forced_variables is not None:
                self.forced_variables = forced_variables[:]
            else:
                self.forced_variables = None
            if limit_file is not None:
                import pickle as pkl
                self.limit_file = pkl.load(open(limit_file, "rb"))
            else:
                self.limit_file = None

    def run(self):
        from AutoRPE.UtilsWorkflow.WorkflowManagment import BinarySearchJob, Queue
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint, recover_checkpoint
        import time

        vault = self.vault
        # Keeping the variables that have an id (it's precision can be changed in the code)
        variables = [v for v in vault.variables if v.id is not None]

        if self.analysis_variables:
            variables = [v for v in variables if v.id in self.analysis_variables]

        # def shift(l, n):
        #     return l[n:] + l[:n]
        #
        # variables = shift(variables, 15)
        print("Starting analysis with %i variables" % len(variables))
        # Original set including all the variables
        original_set = [v.id for v in variables]

        recovering_from_checkpoint = self.restarting_from_checkpoint
        if not recovering_from_checkpoint:
            # Creating the first set that will include all the variables
            whole_set = BinarySearchJob(analysis_set=original_set, forced_variables=self.forced_variables,
                                        reduced_precision_level=self.reduced_precision_level,
                                        communicator=self.communicator, vault=vault, template=self.job_template,
                                        assertion_template=self.assertion_template, limit_file=self.limit_file,
                                        local_folder=self.local_folder)
            whole_set.thresholds = self.error_thresholds
            # Defining it's level to 0
            whole_set.level = 0
            # Initializing an empty queue for the execution
            queue = Queue()
            # Add the whole set to the queue
            queue.pending.append(whole_set)
            # Init loop counter
            loops = 0
        else:
            queue, whole_set, loops = recover_checkpoint(communicator=self.communicator, vault=vault)
            whole_set.vault = vault
            if self.assertion_template:
                for job in queue.all():
                    if job.assertion_job is not None:
                        job.assertion_job.communicator = self.communicator

        # Make a double-precision run to be sure that the accuracy test works as expected
        check_job = BinarySearchJob(
            analysis_set=[],
            forced_variables=self.forced_variables,
            reduced_precision_level=self.reduced_precision_level,
            communicator=self.communicator,
            vault=vault,
            template=self.job_template,
            assertion_template=self.assertion_template,
            limit_file=self.limit_file,
            local_folder=self.local_folder
        )
        check_job.submit()
        while check_job.remote_status != "COMPLETED" and check_job.remote_status != "FAILED":
            time.sleep(5)
        if not check_job.evaluate_success():
            raise AssertionError("Basic simulation to check that the accuracy test works properly didn't pass")
        else:
            print("Basic test was successful, continuing with the analysis")

        # Define a condition to keep iterating with the refinement
        def condition(_job):
            return _job.status is not "SUCCESS" and _job.status is not "FAILED"
        # Start simulations
        wait = False
        waiting_time = 10
        # Counter for using the checkpoint
        change_counter = 0
        while condition(whole_set):
            # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
            loop_start_time = time.time()
            # Main loop
            print("ITERATION: %4i" % loops)
            for key in queue._Queue__dict.keys():
                print("\t%18s: %4i" % (key, len(queue[key])))
            loops += 1

            # Sorting the queue to launch first the lower levels (bigger groups).
            queue.sort(lower_first=True)

            for job in queue.all():
                job_changes = job.driver(queue=queue, max_running_jobs=self.max_running_jobs)
                if job_changes:
                    change_counter += 1
            if change_counter > 10:
                checkpoint(queue=queue, iteration=loops)
                change_counter = 0
            loop_end_time = time.time()
            if wait:
                while loop_end_time - loop_start_time < waiting_time:
                    time.sleep(1)
                    loop_end_time = time.time()

        # Save list of variables
        from AutoRPE.UtilsRPE.Functions import save_variable_list
        save_variable_list(whole_set.variable_set, "%s_final_set.txt" % self.experiment_name, vault=self.vault)
        save_variable_list(whole_set.banned_variables, "%s_banned_variables.txt" % self.experiment_name, vault=self.vault)

        # Saving a last checkpoint
        checkpoint(queue, 1000, filename="end_checkpoint.pkl")